﻿var colors = require('colors');
var pad = require('pad');
var path = require('path');
var Promise = require("bluebird");
var fs = Promise.promisifyAll(require("fs"));
var ping = require('ping-bluebird');
var Dict = require("collections/dict");
var mkdirp = require("mkdirp-then");
var config = require('./actualizar.json');

var ficherosEnUso = []; // Array donde se registran los ficheros en uso.
var hostsDestino = new Dict(); // Diccionario <host, bool> que almacena para cada servidor destino el estado de conexión.

// Obtiene el agrumento que identifica la aplicación a actualizar. Si no se especifica actualiza todos los
// que aparezcan en el fichero de configuración.
var appParaActualizar = null;
if (process.argv.length > 0) {
	var arg0 = process.argv[0].toLowerCase();
    // Si se ejecuta con node.exe, el segundo parámetro es el nombre del .js y el tercero el nombre de la aplicación a actualizar.
    if (arg0.indexOf("node") > -1 && process.argv.length > 2) appParaActualizar = process.argv[2];
    // Si se ejecuta utilizando el .js empaquetado en un ejecutable con JX, el segundo parámetro es el nombre de la aplicación a actualizar.
    else if (arg0.indexOf("node") == -1 && process.argv.length > 1)  appParaActualizar = process.argv[1];
}
if (appParaActualizar != null) {
    console.log("Actualizando " + appParaActualizar);
} else {
    console.log("Actualizando todas las aplicaciones.");
}

// Comprueba la conectividad de los destinos.
Promise.map(Object.keys(config), function (app) {
    if (appParaActualizar == null || appParaActualizar == app) {
        console.log("\nComprobando la conectividad de los destinos para %s:".black.bgWhite, app);
        var destinos = [];
        var destinosApp = config[app]['destinos'];
        for (var key in destinosApp) {
            var destino = destinosApp[key].toLowerCase();
            // Se comprueba la conectividad de los servidores de los destinos. Los destinos locales no los comprueba (P.ej: c:)
            if (destino.indexOf(":") == -1) {
                var host = destino.replace("\\\\", "").split('\\')[0];
                destinos.push(destino);
                if (!hostsDestino.has(host)) {
                    hostsDestino.add(false, host);
                }
            }
        }
        // Realiza un Ping de forma asíncrona para cada servidor destino.
        return Promise.map(hostsDestino.keys(), function (host) {
            return ping(host, {
                timeout: 1
            }).then(function (res) {
                var respuesta = res.alive ? " conectado ".green : " desconectado ".red;
                console.log(pad(res.host, 15) + " " + respuesta);
                hostsDestino.set(res.host, res.alive);
            }).reflect();
        });
    }
}, {concurrency: 20})
    .then(function (object) {
        // Copia a los destinos que tienen conectividad.
        return Promise.map(Object.keys(config), function (app) {
            if (appParaActualizar == null || appParaActualizar == app) {
                console.log("\nIniciando la copia para %s:".black.bgWhite, app);
                var destinos = []; // Array de rutas de los destinos
                var destinosApp = config[app]['destinos'];
                for (var key in destinosApp) {
                    var destino = destinosApp[key].toLowerCase();
                    // Si es una copia local añade el destino. P.e: c:\...
                    if (destino.indexOf(':') > -1) {
                        destinos.push(destino);
                    } else {
                        // Añade la ruta del destino si hay conectividad con el servidor de destino.
                        var host = destino.replace("\\\\", "").split('\\')[0];

                        if (hostsDestino.get(host, false)) {
                            destinos.push(destino);
                        }
                    }
                }
                if (destinos.length > 0) {
                    // Lanza el proceso de copia de la applicación a los destinos.
                    return copiarDirectorio(config[app].origen, config[app].origen, destinos);
                }
            }
        }, {
            concurrency: 100
        });
    }).then(function () {

        console.log("\nResultado:".black.bgWhite);

        // Imprime la información sobre la conectividad de los destinos de las copias.
        if (hostsDestino.values().some(function (conectado, index) {
                return conectado;
            })) {
            console.log("\nDestinos con conectividad:".bold);
            hostsDestino.forEach(function (conectado, servidor) {
                if (conectado)
                    console.log(pad(" ", 4) + "%s", servidor);
            });
        }
        if (hostsDestino.values().some(function (conectado, index) {
                return !conectado;
            })) {
            console.error("\nDestinos sin conectividad:".bold);
            hostsDestino.forEach(function (conectado, servidor) {
                if (!conectado)
                    console.error(pad(" ", 4) + "%s", servidor);
            });
        }

        // Muestra los ficheros que no se han copiado debido a que estaban en uso.
        if (ficherosEnUso.length > 0) {
            console.error("\nFicheros en uso:".bold);
            ficherosEnUso.forEach(function (ruta) {
                console.error(pad(" ", 4) + "%s", ruta);
            });
        }
    });

// Funciones auxiliares
/**
 * Función que lee el contenido de un directorio origen recursivamente y lo copia en uno o varios destinos, todo de forma asíncrona.
 * @param directorio Directorio actual cuyos ficheros se copian a los destinos. Inicialemnte es el mismo que el origen. Conforme va
 recorriendo recursivamente los subdirectorios este parámetro cambia indicando el directorio actual que esta copiando.
 * @param origen Directorio raiz de la aplicación que se copia. No varia en cada iteración recursiva.
 * @param destinos Array de rutas de las carpetas de destino. No varia en cada iteración recursiva.
 * @returns {Function}
 */
function copiarDirectorio(directorio, origen, destinos) {
    // Lee el contenido del directorio. Por cada fichero encontrado ejecuta lo siguiente.
    return fs.readdirAsync(directorio).map(function (fichero) {
        var ruta = path.join(directorio, fichero);
        // Obtiene la información del fichero/directorio
        return fs.statAsync(ruta).then(function (informaciónFichero) {
            // Si es directorio lo copia recursivamente.
            if (informaciónFichero.isDirectory())
                return copiarDirectorio(ruta, origen, destinos);
            else {
                // Si es fichero lo carga en memoria.
                console.log(pad("Leyendo", 15).yellow + ruta.replace(origen, '').gray);
                return fs.readFileAsync(ruta).catch(function ignore(err) {
                })
                    .then(function (contenido) {
                        // Escribe de forma asíncrona el fichero actual en cada destino.
                        return Promise.map(destinos, function (prefijoDestino, index, length) {
                            // Anida la ruta relativa del fichero que esta copiando a la ruta absoluta de la raiz del destino.
                            var destino = path.join(prefijoDestino, ruta.replace(origen, ''));
                            // Crea el directorio de destino si no existe.
                            return mkdirp(path.dirname(destino)).then(function () {
                                console.log(pad("Escribiendo", 15).green + destino);
                                // Escribe el fichero cargado en memoria en el destino.
                                return fs.writeFileAsync(destino, contenido).then(
                                    function onFulfilled(value) {
                                        return {
                                            state: 'fulfilled',
                                            value: value
                                        };
                                    },
                                    function onRejected(reason) {
                                        if (reason.code == "EBUSY") {
                                            ficherosEnUso.push(reason.path);
                                        } else {
                                            console.error("%s", reason.message);
                                        }
                                        // Ignora el rechazo devolviendo fulfilled.
                                        return {
                                            state: 'fulfilled',
                                            value: false
                                        };
                                    });
                            }, function onRejected(reason) {
                                console.error("%s no ha podido ser creada", reason.path);
                                // Ignora el rechazo devolviendo fulfilled.
                                return {
                                    state: 'fulfilled',
                                    value: false
                                };
                            });
                        });
                    }, function onRejected(reason) {
                        console.error("%s no ha podido ser leido." + reason.path);
                        // Ignora el rechazo devolviendo fulfilled.
                        return {
                            state: 'fulfilled',
                            value: false
                        };
                    });
            }
        });
    }, {concurrency: 20});
}

process.on('uncaughtException', function (err) {
    console.error((new Date).toUTCString() + ' Excepción no controlada: '.red, err.message);
});

process.on('unhandledRejection', function (reason, p) {
    if (reason.code == "EBUSY") {
        console.error(reason.path + " está en uso.");
    } else {
        console.error("Error: " + reason.message);
    }
});
